from abc import ABC, abstractmethod

from assimilator.core.database.repository import BaseRepository


class UnitOfWork(ABC):
    def __init__(self, repository: BaseRepository):
        self.repository = repository

    @abstractmethod
    def begin(self):
        raise NotImplementedError()

    @abstractmethod
    def rollback(self):
        raise NotImplementedError()

    @abstractmethod
    def commit(self):
        raise NotImplementedError()

    @abstractmethod
    def close(self):
        raise NotImplementedError()

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.rollback()
            self.close()
            raise exc_val
        else:
            self.close()



from sqlalchemy.exc import NoResultFound, SQLAlchemyError
from sqlalchemy.orm import Query

from assimilator.core.database import BaseRepository, Specification
from assimilator.core.database.exceptions import NotFoundError, DataLayerError


class AlchemyRepository(BaseRepository):
    def __init__(self, session, initial_query: Query = None):
        super(AlchemyRepository, self).__init__(session)
        self.initial_query = initial_query

    def get_initial_query(self):
        if self.initial_query is not None:
            return self.initial_query
        else:
            raise NotImplementedError("You must either pass the initial query "
                                      "to the constructor or define get_initial_query()")

    def _execute_query(self, query):
        return self.session.execute(query)

    def get(self, *specifications: Specification, lazy=False):
        try:
            if lazy:
                return self.apply_specifications(specifications)
            else:
                return self._execute_query(super(AlchemyRepository, self).get(*specifications)).one()[0]
        except NoResultFound as exc:
            raise NotFoundError(exc)

    def filter(self, *specifications: Specification, lazy=False):
        if lazy:
            return self.apply_specifications(specifications)
        else:
            return [result[0] for result in self._execute_query(super(AlchemyRepository, self).filter(*specifications))]

    def update(self, obj):
        """ We don't do anything, as the object is going to be updated with the obj.key = value """

    def update_many(self, specifications, updated_fields):
        try:
            self.filter(*specifications, lazy=True).update(updated_fields)
        except SQLAlchemyError as exc:
            raise DataLayerError(exc)

    def save(self, obj):
        self.session.add(obj)

    def refresh(self, obj):
        self.session.refresh(obj)

    def delete(self, obj):
        self.session.delete(obj)

    def is_modified(self, obj):
        return self.session.is_modified(obj)



from assimilator.core.events import Event
from assimilator.core.events.events_bus import EventConsumer, EventProducer


class InternalEventConsumer(EventConsumer):
    def __init__(self, event_storage: list, listeners: dict):
        self.event_storage = event_storage
        self.listeners = listeners

    def close(self):
        pass

    def start(self):
        pass

    def consume(self):
        for _ in range(len(self.event_storage)):
            event = self.event_storage.pop()

            for listener in self.listeners:
                listener(event)


class InternalEventProducer(EventProducer):
    def __init__(self, event_storage: list):
        self.event_storage = event_storage

    def produce(self, event: Event):
        self.event_storage.append(event)

    def start(self):
        pass

    def close(self):
        pass



from sqlalchemy.orm import Query

from assimilator.core.database.specifications import Specification


class AlchemySpecification(Specification):
    def apply(self, query: Query) -> Query:
        return super(AlchemySpecification, self).apply(query)


class AlchemyFilterSpecification(AlchemySpecification):
    def __init__(self, *filters, **filters_by):
        self.filters = filters
        self.filters_by = filters_by

    def apply(self, query):
        return query.filter(*self.filters).filter_by(**self.filters_by)



from abc import ABC, abstractmethod

from assimilator.core.database.repository import BaseRepository


class DictUnitOfWork(ABC):
    def __init__(self, repository: BaseRepository):
        self.repository = repository
        self._saved_data = None

    @abstractmethod
    def begin(self):
        self._saved_data = self.repository.session
        self.repository.session = dict(self._saved_data)

    @abstractmethod
    def rollback(self):
        self.repository.session = self._saved_data

    @abstractmethod
    def commit(self):
        self._saved_data = self.repository.session

    @abstractmethod
    def close(self):
        pass



from setuptools import setup, find_packages

setup(
    name='py_assimilator',
    version='0.1.5',
    author='Andrey Ivanov',
    author_email='python.on.papyrus@gmail.com',
    url='https://pypi.python.org/pypi/py_assimilator/',
    project_urls={
        'Documentation': 'https://github.com/knucklesuganda/py_assimilator',
        'Funding': 'https://github.com/knucklesuganda/py_assimilator',
        'Say Thanks!': 'https://github.com/knucklesuganda/py_assimilator',
        'Source': 'https://github.com/knucklesuganda/py_assimilator',
        'Tracker': 'https://github.com/knucklesuganda/py_assimilator',
    },
    license='LICENSE.md',
    packages=find_packages(),
    description='The best python patterns for your projects',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    requires=['pydantic'],
    extras_require={
        'database': ['SQLAlchemy'],
        'kafka': ['kafka-python'],
        'full': ['SQLAlchemy', 'kafka-python'],
    },
)



from abc import ABC, abstractmethod

from assimilator.core.events.events import Event
from assimilator.core.patterns.context_managers import StartCloseContextMixin


class EventConsumer(StartCloseContextMixin):
    @abstractmethod
    def consume(self):
        raise NotImplementedError("consume() is not implemented")


class EventProducer(StartCloseContextMixin):
    @abstractmethod
    def produce(self, event: Event):
        raise NotImplementedError("produce() is not implemented")


class EventBus(ABC):
    def __init__(self, consumer: EventConsumer, producer: EventProducer):
        self.consumer = consumer
        self.producer = producer

    @abstractmethod
    def produce(self, event: Event):
        self.producer.produce(event)

    @abstractmethod
    def consume(self):
        return self.consumer.consume()
